unit TabControl;

interface

uses
  SysUtils, Windows, Messages, Types, Classes, Controls, Graphics, Menus;

type
  TSimpleTabControl = class(TCustomControl)
  private const
    DEFAULT_PADDING = 4;
    DEFAULT_TAB_WIDTH = 160;
    DEFAULT_ACTIVE_TAB_COLOR = clWhite;
    DEFAULT_ACTIVE_TAB_TEXT_COLOR = clBlack;
    DEFAULT_BORDER_COLOR = clBlack;
    DEFAULT_INACTIVE_TAB_COLOR = clBtnFace;
    DEFAULT_INACTIVE_TAB_TEXT_COLOR = clBtnText;
  private var
    FItems: TStrings;
    FColor: TColor;
    FCurrentTabWidth: integer;
    FDefaultTabWidth: integer;
    FFont: TFont;
    FActiveTabColor: TColor;
    FInactiveTabColor: TColor;
    FActiveTabTextColor: TColor;
    FInactiveTabTextColor: TColor;
    FActiveTab: integer;
    FPadding: integer;
    FOnChange: TNotifyEvent;
    FClickFocuses: boolean;
    FBorderColor: TColor;
    FTabPopupMenu: TPopupMenu;
    FOverflows: array of boolean;
    procedure ItemsChanged(Sender: TObject);
    procedure SetColor(const Value: TColor);
    procedure SetDefaultTabWidth(const Value: integer);
    procedure SetFont(const Value: TFont);
    procedure FontChanged(Sender: TObject);
    procedure SetActiveTabColor(const Value: TColor);
    procedure SetInactiveTabColor(const Value: TColor);
    procedure SetActiveTabTextColor(const Value: TColor);
    procedure SetInactiveTabTextColor(const Value: TColor);
    procedure SetActiveTab(const Value: integer);
    function GetTabRect(const TabIndex: integer): TRect;
    procedure SetItems(const Value: TStrings);
    procedure SetPadding(const Value: integer);
    procedure DoChange;
    procedure SetBorderColor(const Value: TColor);
    function UpdateTabWidth: boolean;
  protected
    procedure Paint; override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X: Integer;
      Y: Integer); override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure WndProc(var Message: TMessage); override;
    procedure CMEnabledchanged(var Message: TMessage);
      message CM_ENABLEDCHANGED;
    procedure WMContextMenu(var Message: TWMContextMenu);
      message WM_CONTEXTMENU;
    procedure WMSize(var Message: TWMSize); message WM_SIZE;
    procedure CMHintShow(var Message: TCMHintShow); message CM_HINTSHOW;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function GetTabAtPoint(const APoint: TPoint): integer;
  published
    property Align;
    property Anchors;
    property Cursor;
    property Enabled;
    property Margins;
    property ClickFocuses: boolean read FClickFocuses write FClickFocuses default false;
    property BorderColor: TColor read FBorderColor write SetBorderColor default DEFAULT_BORDER_COLOR;
    property Items: TStrings read FItems write SetItems;
    property Color: TColor read FColor write SetColor default clBtnFace;
    property DefaultTabWidth: integer read FDefaultTabWidth write SetDefaultTabWidth default DEFAULT_TAB_WIDTH;
    property Font: TFont read FFont write SetFont;
    property InactiveTabColor: TColor read FInactiveTabColor write SetInactiveTabColor default DEFAULT_INACTIVE_TAB_COLOR;
    property ActiveTabColor: TColor read FActiveTabColor write SetActiveTabColor default DEFAULT_ACTIVE_TAB_COLOR;
    property InactiveTabTextColor: TColor read FInactiveTabTextColor write SetInactiveTabTextColor default DEFAULT_INACTIVE_TAB_TEXT_COLOR;
    property ActiveTabTextColor: TColor read FActiveTabTextColor write SetActiveTabTextColor default DEFAULT_ACTIVE_TAB_TEXT_COLOR;
    property ActiveTab: integer read FActiveTab write SetActiveTab default 0;
    property Padding: integer read FPadding write SetPadding default DEFAULT_PADDING;
    property PopupMenu;
    property TabPopupMenu: TPopupMenu read FTabPopupMenu write FTabPopupMenu;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property TabStop;
    property TabOrder;
    property OnClick;
    property OnMouseDown;
    property OnMouseUp;
    property OnKeyPress;
    property OnKeyDown;
    property OnKeyUp;
  end;

procedure Register;

implementation

uses Math;

procedure Register;
begin
  RegisterComponents('Rejbrand 2015', [TSimpleTabControl]);
end;

{ TSimpleTabControl }

procedure TSimpleTabControl.CMEnabledchanged(var Message: TMessage);
begin
  Invalidate;
end;

procedure TSimpleTabControl.CMHintShow(var Message: TCMHintShow);
var
  TabIndex: integer;
begin
  TabIndex := GetTabAtPoint(Message.HintInfo^.CursorPos);
  if InRange(TabIndex, 0, FItems.Count - 1) and InRange(TabIndex, low(FOverflows), high(FOverflows)) then
  begin
    if FOverflows[TabIndex] then
    begin
      Message.HintInfo^.CursorRect := GetTabRect(TabIndex);
      Message.HintInfo^.HintStr := FItems[TabIndex];
      Message.HintInfo^.HideTimeout := 60000;
    end;
  end;
end;

constructor TSimpleTabControl.Create(AOwner: TComponent);
begin
  inherited;
  Width := 400;
  Height := 32;
  FClickFocuses := false;
  FActiveTab := 0;
  FPadding := DEFAULT_PADDING;
  FBorderColor := DEFAULT_BORDER_COLOR;
  FActiveTabColor := DEFAULT_ACTIVE_TAB_COLOR;
  FInactiveTabColor := DEFAULT_INACTIVE_TAB_COLOR;
  FActiveTabTextColor := DEFAULT_ACTIVE_TAB_TEXT_COLOR;
  FInactiveTabTextColor := DEFAULT_INACTIVE_TAB_TEXT_COLOR;
  FDefaultTabWidth := DEFAULT_TAB_WIDTH;
  FColor := clBtnFace;
  FFont := TFont.Create;
  FFont.OnChange := FontChanged;
  FItems := TStringList.Create;
  TStringList(FItems).OnChange := ItemsChanged;
  ShowHint := true;
end;

destructor TSimpleTabControl.Destroy;
begin
  FFont.Free;
  FItems.Free;
  inherited;
end;

procedure TSimpleTabControl.DoChange;
begin
  if Assigned(FOnChange) then
    FOnChange(Self);
end;

procedure TSimpleTabControl.FontChanged(Sender: TObject);
begin
  Invalidate;
end;

function TSimpleTabControl.GetTabAtPoint(const APoint: TPoint): integer;
begin
  result := APoint.X div FCurrentTabWidth;
  if not InRange(result, 0, FItems.Count - 1) then
    result := -1;
end;

function TSimpleTabControl.GetTabRect(const TabIndex: integer): TRect;
begin
  result.Left := TabIndex * FCurrentTabWidth;
  result.Right := result.Left + FCurrentTabWidth;
  result.Top := 0;
  result.Bottom := ClientHeight;
end;

procedure TSimpleTabControl.ItemsChanged(Sender: TObject);
begin
  UpdateTabWidth;
  Invalidate;
end;

procedure TSimpleTabControl.KeyDown(var Key: Word; Shift: TShiftState);
begin
  inherited;
  case Key of
    VK_RIGHT:
      if FActiveTab < FItems.Count - 1 then
      begin
        SetActiveTab(FActiveTab + 1);
        DoChange;
      end;
    VK_LEFT:
      if FActiveTab > 0 then
      begin
        SetActiveTab(FActiveTab - 1);
        DoChange;
      end;
    VK_HOME:
      if FActiveTab <> 0 then
      begin
        SetActiveTab(0);
        DoChange;
      end;
    VK_END:
      if FActiveTab <> FItems.Count - 1 then
      begin
        SetActiveTab(FItems.Count - 1);
        DoChange;
      end;
  end;
end;

procedure TSimpleTabControl.MouseDown(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
var
  tab: integer;
begin
  if FClickFocuses and CanFocus then SetFocus;
  tab := GetTabAtPoint(Point(X, Y));
  if (tab <> -1) and (tab <> FActiveTab) then
  begin
    SetActiveTab(tab);
    DoChange;
  end;
  inherited;
end;

procedure TSimpleTabControl.Paint;
var
  i: Integer;
  R, R2: TRect;
  S: string;
  h: integer;
begin
  inherited;
  SetLength(FOverflows, FItems.Count);

  Canvas.Brush.Color := FColor;
  Canvas.FillRect(ClientRect);
  Canvas.Font.Assign(FFont);

  // Backgrounds
  for i := 0 to FItems.Count - 1 do
  begin
    R := GetTabRect(i);
    S := FItems[i];
    if FActiveTab = i then
      Canvas.Font.Style := [fsBold]
    else
      Canvas.Font.Style := [];
    if not Enabled then
    begin
      Canvas.Brush.Color := clBtnFace;
      Canvas.Font.Color := clBtnText;
    end
    else if FActiveTab = i then
    begin
      Canvas.Brush.Color := FActiveTabColor;
      Canvas.Font.Color := FActiveTabTextColor;
    end
    else
    begin
      Canvas.Brush.Color := FInactiveTabColor;
      Canvas.Font.Color := FInactiveTabTextColor;
    end;
    Canvas.FillRect(R);
    InflateRect(R, -FPadding, 0);
    DrawText(Canvas.Handle, PChar(S), Length(S),
      R, DT_SINGLELINE or DT_VCENTER or DT_LEFT or DT_END_ELLIPSIS or DT_NOPREFIX);
    R2 := R;
    DrawText(Canvas.Handle, PChar(S), Length(S),
      R2, DT_SINGLELINE or DT_VCENTER or DT_LEFT or DT_NOPREFIX or DT_CALCRECT);
    FOverflows[i] := R2.Right > R.Right;
    if Focused and (FActiveTab = i) then
    begin
      h := DrawText(Canvas.Handle, PChar(S), Length(S), R,
        DT_SINGLELINE or DT_VCENTER or DT_LEFT or DT_END_ELLIPSIS or DT_NOPREFIX or DT_CALCRECT);
      R.Top := ClientHeight div 2 - h div 2;
      R.Bottom := R.Top + h;
      InflateRect(R, 2, 0);
      Canvas.DrawFocusRect(R);
    end;
  end;

  // Borders
  for i := 0 to FItems.Count - 1 do
  begin
    R := GetTabRect(i);
    Canvas.Pen.Color := FBorderColor;
    if FActiveTab = i then
    begin
      Canvas.MoveTo(R.Left, R.Bottom);
      Canvas.LineTo(R.Left, R.Top);
      Canvas.LineTo(R.Right, R.Top);
      Canvas.LineTo(R.Right, R.Bottom);
    end
    else
    begin
      Canvas.MoveTo(R.Left, R.Bottom - 1);
      Canvas.LineTo(R.Right, R.Bottom -  1);
    end;
    if i = FItems.Count - 1 then
    begin
      Canvas.MoveTo(R.Right, R.Bottom - 1);
      Canvas.LineTo(ClientWidth, R.Bottom - 1);
    end;
    if (i > 0) and (FActiveTab <> i - 1) and (FActiveTab <> i) then
    begin
      Canvas.Pen.Color := FColor;
      Canvas.MoveTo(R.Left, R.Top);
      Canvas.LineTo(R.Left, R.Bottom - 1);
    end;
  end;
end;

procedure TSimpleTabControl.SetActiveTab(const Value: integer);
begin
  if FActiveTab <> Value then
  begin
    FActiveTab := Value;
    Invalidate;
  end;
end;

procedure TSimpleTabControl.SetActiveTabColor(const Value: TColor);
begin
  if FActiveTabColor <> Value then
  begin
    FActiveTabColor := Value;
    Invalidate;
  end;
end;

procedure TSimpleTabControl.SetActiveTabTextColor(const Value: TColor);
begin
  if FActiveTabTextColor <> Value then
  begin
    FActiveTabTextColor := Value;
    Invalidate;
  end;
end;

procedure TSimpleTabControl.SetBorderColor(const Value: TColor);
begin
  if FBorderColor <> Value then
  begin
    FBorderColor := Value;
    Invalidate;
  end;
end;

procedure TSimpleTabControl.SetColor(const Value: TColor);
begin
  if FColor <> Value then
  begin
    FColor := Value;
    Invalidate;
  end;
end;

procedure TSimpleTabControl.SetFont(const Value: TFont);
begin
  if FFont <> Value then
  begin
    FFont.Assign(Value);
    Invalidate;
  end;
end;

procedure TSimpleTabControl.SetInactiveTabColor(const Value: TColor);
begin
  if FInactiveTabColor <> Value then
  begin
    FInactiveTabColor := Value;
    Invalidate;
  end;
end;

procedure TSimpleTabControl.SetInactiveTabTextColor(const Value: TColor);
begin
  if FInactiveTabTextColor <> Value then
  begin
    FInactiveTabTextColor := Value;
    Invalidate;
  end;
end;

procedure TSimpleTabControl.SetItems(const Value: TStrings);
begin
  if FItems <> Value then
  begin
    FItems.Assign(Value);
    Invalidate;
  end;
end;

procedure TSimpleTabControl.SetPadding(const Value: integer);
begin
  if FPadding <> Value then
  begin
    FPadding := Value;
    Invalidate;
  end;
end;

procedure TSimpleTabControl.SetDefaultTabWidth(const Value: integer);
begin
  if FDefaultTabWidth <> Value then
  begin
    FDefaultTabWidth := Value;
    if UpdateTabWidth then
      Invalidate;
  end;
end;

function TSimpleTabControl.UpdateTabWidth: boolean;
var
  OldTabWidth: integer;
begin
  OldTabWidth := FCurrentTabWidth;
  if FItems.Count > 0 then
    FCurrentTabWidth := Max(16, Min(FDefaultTabWidth, ClientWidth div FItems.Count))
  else
    FCurrentTabWidth := FDefaultTabWidth;
  result := OldTabWidth <> FCurrentTabWidth;
end;

procedure TSimpleTabControl.WMContextMenu(var Message: TWMContextMenu);
begin
  if (Message.XPos = -1) and (Message.YPos = -1) then // menu key of Shift+F10
  begin
    if Assigned(FTabPopupMenu) and Visible and (Items.Count > 0) then
      with ClientToScreen(Point(FActiveTab * FCurrentTabWidth, ClientHeight - 1)) do
        FTabPopupMenu.Popup(X, Y);
  end
  else
    if Assigned(FTabPopupMenu) and (GetTabAtPoint(ScreenToClient(SmallPointToPoint(Message.Pos))) <> -1) then
      with SmallPointToPoint(Message.Pos) do
        FTabPopupMenu.Popup(X, Y)
    else
      inherited;
end;

procedure TSimpleTabControl.WMSize(var Message: TWMSize);
begin
  inherited;
  if UpdateTabWidth then
    Invalidate;
end;

procedure TSimpleTabControl.WndProc(var Message: TMessage);
begin
  inherited;
  case Message.Msg of
    WM_GETDLGCODE:
      Message.Result := Message.Result or DLGC_WANTARROWS;
    WM_SETFOCUS, WM_KILLFOCUS:
      Invalidate;
  end;
end;

end.
